# pylint: disable=unnecessary-comprehension, arguments-differ, attribute-defined-outside-init, disallowed-name
# pylint: disable=too-many-locals, too-many-statements

from django.test.testcases import (
    TestCase,
)

from edureporter import (
    AbstractDataProvider,
    CompositeDataProvider,
    DependentCompositeProvider,
    FlatDataProviderAdapter,
    NestedDataProviderAdapter,
)


class Provider01(AbstractDataProvider):
    """Тестовый провайдер #1."""

    def init(self, a, b, c):
        self._a = a
        self._b = b
        self._c = c

    @property
    def param01(self):
        return self._a

    @property
    def param02(self):
        return self._b

    @property
    def param03(self):
        return self._c


class Provider02(AbstractDataProvider):
    """Тестовый провайдер #2."""

    def init(self, a, b, c):
        self._a = a
        self._b = b
        self._c = c

    @property
    def param01(self):
        return self._a

    @property
    def param02(self):
        return self._b

    @property
    def param03(self):
        return self._c

    @property
    def param04(self):
        return self._a + self._b + self._c


class Provider03(AbstractDataProvider):
    """Тестовый провайдер #3."""

    def init(self, a, b, c):
        self._a = a
        self._b = b
        self._c = c

    @property
    def param01(self):
        return self._a

    @property
    def param02(self):
        return self._b

    @property
    def param03(self):
        return self._c


class Provider04(AbstractDataProvider):
    """Тестовый провайдер #4."""

    def init(self, a, b, c):
        self._a = a
        self._b = b
        self._c = c

    @property
    def param01(self):
        return self._a

    @property
    def param02(self):
        return self._b

    @property
    def param03(self):
        return self._c


# значения для теста свойств провайдеров
A, B, C = 1, 2, 3
D = A + B + C
BAR, FOO = 42, 2128506

# число параметров во всём дереве провайдеров
PARAMS_COUNT = 15


class MainDataProvider(CompositeDataProvider):
    """Композитный тестовый провайдер #1."""

    provider_name = 'Main'

    @property
    def bar(self):
        return BAR


class SuperDataProvider(CompositeDataProvider):
    """Композитный тестовый провайдер #2."""

    @property
    def foo(self):
        return FOO


def setup_provider():

    sub_provider = MainDataProvider([
        Provider01,
        Provider02,
    ])

    provider = SuperDataProvider([
        sub_provider,
        Provider03,
        Provider04,
    ])

    provider.init(a=A, b=B, c=C)
    provider.load_data()

    return provider


class FlatAdapterTestCase(TestCase):
    """Тестирование "плоского" адаптера."""

    def setUp(self):
        provider = setup_provider()
        self.adapter = FlatDataProviderAdapter(provider, splitter='__')

    def test_common(self):
        adapter = self.adapter
        self.assertTrue(adapter)
        self.assertEqual(len(self.adapter), PARAMS_COUNT)
        self.assertEqual(adapter.get('foo'), FOO)
        self.assertEqual(adapter.get('foo42', 42), 42)

    def test_contains(self):
        adapter = self.adapter
        self.assertIn('Provider03__param01', adapter)
        self.assertIn('Provider03__param02', adapter)
        self.assertIn('Provider03__param03', adapter)
        self.assertIn('Provider04__param01', adapter)
        self.assertIn('Provider04__param02', adapter)
        self.assertIn('Provider04__param03', adapter)
        self.assertIn('Main__Provider01__param01', adapter)
        self.assertIn('Main__Provider01__param02', adapter)
        self.assertIn('Main__Provider01__param03', adapter)
        self.assertIn('Main__Provider02__param01', adapter)
        self.assertIn('Main__Provider02__param02', adapter)
        self.assertIn('Main__Provider02__param03', adapter)
        self.assertIn('Main__bar', adapter)
        self.assertIn('foo', adapter)

        self.assertEqual(adapter['Provider03__param01'], A)
        self.assertEqual(adapter['Provider03__param02'], B)
        self.assertEqual(adapter['Provider03__param03'], C)
        self.assertEqual(adapter['Provider04__param01'], A)
        self.assertEqual(adapter['Provider04__param02'], B)
        self.assertEqual(adapter['Provider04__param03'], C)
        self.assertEqual(adapter['Main__Provider01__param01'], A)
        self.assertEqual(adapter['Main__Provider01__param02'], B)
        self.assertEqual(adapter['Main__Provider01__param03'], C)
        self.assertEqual(adapter['Main__Provider02__param01'], A)
        self.assertEqual(adapter['Main__Provider02__param02'], B)
        self.assertEqual(adapter['Main__Provider02__param03'], C)
        self.assertEqual(adapter['Main__Provider02__param04'], D)
        self.assertEqual(adapter['Main__bar'], BAR)
        self.assertEqual(adapter['foo'], FOO)

    def test_iter(self):
        adapter = self.adapter

        keys = list(adapter.keys())
        iter_keys = [k for k in adapter]
        self.assertEqual(keys, iter_keys)
        self.assertEqual(len(keys), PARAMS_COUNT)

        iteritems = [x for x in adapter.items()]
        self.assertEqual(len(iteritems), PARAMS_COUNT)
        for key, value in iteritems:
            self.assertEqual(value, adapter[key])

        items = list(adapter.items())
        for (key, value) in items:
            self.assertEqual(value, adapter[key])

        values = list(adapter.values())
        iter_values = [v for v in adapter.values()]
        self.assertEqual(values, iter_values)
        self.assertEqual(len(values), PARAMS_COUNT)


class NestedAdapterTestCase(TestCase):
    """Тестирование иерархического адаптера."""

    def setUp(self):
        provider = setup_provider()
        self.adapter = NestedDataProviderAdapter(provider)

    def test_common(self):
        adapter = self.adapter
        self.assertTrue(adapter)
        self.assertEqual(len(adapter), 4)
        self.assertEqual(len(adapter['Main']), 3)
        self.assertEqual(len(adapter['Main']['Provider01']), 3)
        self.assertEqual(len(adapter['Main']['Provider02']), 4)
        self.assertEqual(len(adapter['Provider03']), 3)
        self.assertEqual(len(adapter['Provider04']), 3)
        self.assertEqual(adapter.get('foo'), FOO)
        self.assertEqual(adapter.get('foo42', 42), 42)

    def test_contains(self):
        adapter = self.adapter
        self.assertIn('foo', adapter)
        self.assertIn('Main', adapter)
        self.assertIn('Provider01', adapter['Main'])
        self.assertIn('Provider02', adapter['Main'])
        self.assertIn('Provider03', adapter)
        self.assertIn('Provider04', adapter)

        self.assertEqual(adapter['Provider03']['param01'], A)
        self.assertEqual(adapter['Provider03']['param02'], B)
        self.assertEqual(adapter['Provider03']['param03'], C)
        self.assertEqual(adapter['Provider04']['param01'], A)
        self.assertEqual(adapter['Provider04']['param02'], B)
        self.assertEqual(adapter['Provider04']['param03'], C)
        self.assertEqual(adapter['Main']['Provider01']['param01'], A)
        self.assertEqual(adapter['Main']['Provider01']['param02'], B)
        self.assertEqual(adapter['Main']['Provider01']['param03'], C)
        self.assertEqual(adapter['Main']['Provider02']['param01'], A)
        self.assertEqual(adapter['Main']['Provider02']['param02'], B)
        self.assertEqual(adapter['Main']['Provider02']['param03'], C)
        self.assertEqual(adapter['Main']['Provider02']['param04'], D)
        self.assertEqual(adapter['Main']['bar'], BAR)
        self.assertEqual(adapter['foo'], FOO)

    def test_iter(self):
        adapter = self.adapter

        keys = list(adapter.keys())
        iter_keys = [k for k in adapter]
        self.assertEqual(keys, iter_keys)
        self.assertEqual(len(keys), 4)

        main_keys = list(adapter['Main'].keys())
        main_iter_keys = [k for k in adapter['Main']]
        self.assertEqual(main_keys, main_iter_keys)
        self.assertEqual(len(main_keys), 3)

        p3_keys = list(adapter['Provider03'].keys())
        p3_iter_keys = [k for k in adapter['Provider03']]
        self.assertEqual(p3_keys, p3_iter_keys)
        self.assertEqual(len(p3_keys), 3)

        p4_keys = list(adapter['Provider04'].keys())
        p4_iter_keys = [k for k in adapter['Provider04']]
        self.assertEqual(p4_keys, p4_iter_keys)
        self.assertEqual(len(p4_keys), 3)

        iteritems = [x for x in adapter.items()]
        self.assertEqual(len(iteritems), 4)
        for key, value in iteritems:
            self.assertEqual(value, adapter[key])

        items = list(adapter.items())
        self.assertEqual(len(items), 4)
        for key, value in items:
            self.assertEqual(value, adapter[key])

        main_iteritems = [x for x in adapter['Main'].items()]
        self.assertEqual(len(main_iteritems), 3)
        for key, value in main_iteritems:
            self.assertEqual(value, adapter['Main'][key])

        main_items = list(adapter['Main'].items())
        self.assertEqual(len(main_items), 3)
        for key, value in main_items:
            self.assertEqual(value, adapter['Main'][key])

        p3_iteritems = [x for x in adapter['Provider03'].items()]
        self.assertEqual(len(p3_iteritems), 3)
        for key, value in p3_iteritems:
            self.assertEqual(value, adapter['Provider03'][key])

        p3_items = [x for x in adapter['Provider03'].items()]
        self.assertEqual(len(p3_items), 3)
        for key, value in p3_items:
            self.assertEqual(value, adapter['Provider03'][key])

        p4_iteritems = [x for x in adapter['Provider04'].items()]
        self.assertEqual(len(p4_iteritems), 3)
        for key, value in p4_iteritems:
            self.assertEqual(value, adapter['Provider04'][key])

        p4_items = [x for x in adapter['Provider04'].items()]
        self.assertEqual(len(p4_items), 3)
        for key, value in p4_items:
            self.assertEqual(value, adapter['Provider04'][key])

        values = list(adapter.values())
        iter_values = [v for v in adapter.values()]
        self.assertEqual(values, iter_values)
        self.assertEqual(len(values), 4)

        main_values = list(adapter['Main'].values())
        main_iter_values = [k for k in adapter['Main'].values()]
        self.assertEqual(main_values, main_iter_values)
        self.assertEqual(len(main_values), 3)

        p1_values = list(adapter['Main']['Provider01'].values())
        p1_iter_values = [
            k for k in adapter['Main']['Provider01'].values()
        ]
        self.assertEqual(p1_values, p1_iter_values)
        self.assertEqual(len(p1_values), 3)

        p2_values = list(adapter['Main']['Provider02'].values())
        p2_iter_values = [
            k for k in adapter['Main']['Provider02'].values()
        ]
        self.assertEqual(p2_values, p2_iter_values)
        self.assertEqual(len(p2_values), 4)

        p3_values = list(adapter['Provider03'].values())
        p3_iter_values = [k for k in adapter['Provider03'].values()]
        self.assertEqual(p3_values, p3_iter_values)
        self.assertEqual(len(p3_values), 3)

        p4_values = list(adapter['Provider04'].values())
        p4_iter_values = [k for k in adapter['Provider04'].values()]
        self.assertEqual(p4_values, p4_iter_values)
        self.assertEqual(len(p4_values), 3)


# -----------------------------------------------------------------------------
# Тест зависимого провайдера

class LowLevel01(AbstractDataProvider):
    """Простой подпродвайдер."""

    data = None  # данные будут здесь

    def init(self, a, b, c, **params):
        self._a = a
        self._b = b
        self._c = c

    def load_data(self):
        self.data = self._a + self._b + self._c


class LowLevel02(AbstractDataProvider):
    """Простой подпродвайдер."""

    data = None  # данные будут здесь

    def init(self, d, e, **params):
        self._d = d
        self._e = e

    def load_data(self):
        self.data = self._d + self._e


class LowLevelDependent03(AbstractDataProvider):
    """
    Простой подпродвайдер.

    Зависит от данных LowLevel01.data и LowLevel02.data
    """

    def init(self, f, **params):
        self._f = f
        # придет снаружи
        self.a_b_c = None
        self.d_e = None

    def load_data(self):
        self.data = self.a_b_c + self.d_e + self._f


class HighLevelProvider(DependentCompositeProvider):
    """Композитный провайдер с описанием подпровайдеров."""

    providers_order = (
        ('_low_level_01', LowLevel01),
        ('_low_level_02', LowLevel02),
        ('_low_level_03', LowLevelDependent03)
    )

    _dependence_map = {
        '_low_level_03': {
            'a_b_c': ('_low_level_01', 'data'),
            'd_e': ('_low_level_02', 'data'),
        }
    }

    def init(self, g, **params):
        super().init(**params)
        self._g = g

    def load_data(self):
        super().load_data()
        self.data = self._g + self._low_level_03.data


class DependentProviderTestCase(TestCase):
    """Тестирование композитного провайдера с описанием подпровайдеров."""

    def setUp(self):
        self._initial_data = {
            'a': 1,
            'b': 2,
            'c': 3,
            'd': 4,
            'e': 5,
            'f': 6,
            'g': 7,
        }
        self.provider = HighLevelProvider()
        self.provider.init(**self._initial_data)

    def test_sum(self):
        """
        Сверка суммы.

        Подпровайдеры занимались суммой переданных им значений.
        Один из подпровайдеров (LowLevelDependent03) - зависим от данных
        LowLevel01 и LowLevel02
        """
        self.provider.load_data()

        self.assertEqual(
            self.provider.data,
            sum(self._initial_data.values())
        )



# pylint: disable=too-many-arguments, too-many-locals, too-many-positional-arguments
from .fonts import (
    ARIAL,
    get_font,
    split_text,
)


# Высота и ширина символа "0" шрифта по умолчанию в xlwt.
DEFAULT_CHAR_SIZE = 256


def cm_to_inch(value):
    """Переводит значение из сантиметров в дюймы."""
    return value / 2.54


def inch_to_cm(value):
    """Переводит значение из дюймов в сантиметры."""
    return value * 2.54


def get_cell_bounds(section, row_index, column_index):
    """Возвращает границы ячейки, находящейся на пересечении строки и столбца.

    .. note::

       Индексы строк и колонок начинаются с нуля.

    :param section: Секция, в которой находится строка.
    :type section: :class:`simple_report.xls.section.Section`

    :param int row_index: Индекс строки в таблице excel.

    :param int column_index: Индекс столбца в таблице excel.

    :returns: Кортеж из четырех целых чисел: верхней строки, нижней строки,
        левой колонки и правой колонки.
    :rtype: tuple
    """
    merged_ranges = section.writer.wtsheet.merged_ranges
    for first_row, last_row, first_column, last_column in merged_ranges:
        if (
            first_row <= row_index <= last_row and
            first_column <= column_index <= last_column
        ):
            return first_row, last_row, first_column, last_column

    return row_index, row_index, column_index, column_index


def get_cell_width(section, row_index, column_index):
    """Возвращает ширину ячейки с учётом объединения.

    .. note ::

       Ширина ячейки определяется, как 1/256 от ширины символа "0" первого
       попавшегося шрифта в файле, т.е. единицы изменения достаточно условны.

    .. note::

       Индексы строк и колонок начинаются с нуля.

    :param section: Секция, в которой находится строка.
    :type section: :class:`simple_report.xls.section.Section`

    :param int row_index: Индекс строки в таблице excel.

    :param int column_index: Индекс столбца в таблице excel.

    :rtype: int
    """
    _, _, first_column, last_column = get_cell_bounds(
        section, row_index, column_index
    )
    return sum(
        section.writer.wtsheet.col(i).width
        for i in range(first_column, last_column + 1)
    )


def get_cell_height(section, row_index, column_index):
    """Возвращает высоту ячейки с учётом объединения.

    .. note::

       Индексы строк и колонок начинаются с нуля.

    :param section: Секция, в которой находится строка.
    :type section: :class:`simple_report.xls.section.Section`

    :param int row_index: Индекс строки в таблице excel.

    :param int column_index: Индекс столбца в таблице excel.

    :returns: Высоту ячейки в `твипах <https://goo.gl/hfUW7x>`_.
    :rtype: int
    """
    first_row, last_row, _, _ = get_cell_bounds(
        section, row_index, column_index
    )
    return sum(
        section.writer.wtsheet.row(i).height
        for i in range(first_row, last_row + 1)
    )


def adjust_row_height(section, row_idx, col_idx, text, font,
                      adjusted_row_index=None):
    """Увеличивает высоту строки, если необходимо.

    Высота строки устанавливается такая, чтобы текст `text`
    поместился в ячейку с индексами `row_idx`, `col_idx`.

    Можно применять для нескольких ячеек одной строки. Тогда строке будет
    установлена максимальная высота.

    .. note::

       Индексы строк и колонок начинаются с нуля.

    :param section: Секция, в которой находится строка.
    :type section: :class:`simple_report.xls.section.Section`

    :param int row_idx: Индекс строки в таблице excel.

    :param int col_idx: Индекс столбца в таблице excel.

    :param str text: Строка, которая будет записана в ячейку.

    :param font: Шрифт.
    :type font: :class:`PIL.ImageFont.FreeTypeFont`

    :param int adjusted_row_index: Номер строки, размер которой будет увеличен.
        Если не указывается (``None``), то высота всех строк объединенной
        ячейки будет увеличена равномерно. Номер указывается относительно
        ячейки, нумерация начинается с нуля.
    """
    def get_text_height(column_width):
        """Вычисляет высоту строки отчета в зависимости от текста."""
        # Т.к. высота и ширина ячейки измеряется в единицах относительно
        # дефолтного шрифта, берем его параметры.
        normal_font = get_font(ARIAL, 10)

        # Количество символов "0", которое входит в одну ячейку без переносов.
        # DEFAULT_CHAR_SIZE * 0.8 определяет ширину отступов слева и справа (в
        # сумме, а не каждого отступа по отдельности). Источник точной
        # информации об отступах в ячейке найти не удалось.
        width_in_chars = (
            int(column_width - DEFAULT_CHAR_SIZE * 0.8) // DEFAULT_CHAR_SIZE
        )

        str_width_px, str_height_px = normal_font.getsize('0' * width_in_chars)

        # Определяем коэффициент масштабирования высоты шрифта к дефолтному.
        _, choosen_font_height_px = font.getsize('0')
        height_scale = (choosen_font_height_px * 1.0) / str_height_px

        rows_count = len(
            split_text(text, font, str_width_px)
        )

        height = int(rows_count * height_scale * DEFAULT_CHAR_SIZE)

        return height

    first_row, last_row, first_column, last_column = get_cell_bounds(
        section, row_idx, col_idx
    )

    merged = first_row != last_row or first_column != last_column

    cell_width = get_cell_width(section, row_idx, col_idx)
    cell_height = get_cell_height(section, row_idx, col_idx)

    # Высотя ячейки, в которой поместится указанный текст.
    text_height = get_text_height(cell_width)

    # Разница между текущей высотой ячейки и необходимой для размещения текста
    # высотой.
    height_delta = max(0, text_height - cell_height)

    def add_row_height(row, height):
        row.height_mismatch = True
        row.height += height

    if height_delta > 0:
        if merged:
            row_count = last_row - first_row + 1
            if adjusted_row_index is None:
                # Равномерное увеличение высоты строк
                for r in range(first_row, last_row + 1):
                    row = section.writer.wtsheet.row(r)
                    add_row_height(row, height_delta // row_count)
            else:
                assert adjusted_row_index > row_count, (
                    adjusted_row_index, row_count
                )
                row = section.writer.wtsheet.row(
                    first_row + adjusted_row_index
                )
                add_row_height(row, height_delta)
        else:
            row = section.writer.wtsheet.row(row_idx)
            add_row_height(row, height_delta)


def adjust_row_height_arial(section, row_idx, col_idx, text, font_size=10,
                            adjusted_row_index=None):
    """Устанавливает высоту строки автоматически.

    Высота строки устанавливается такая, чтобы текст text
    поместился в ячейку с индексами row_idx, col_idx.

    Расчет ведется для шрифта Arial, и соответствует отображению ячеек в
    MS Excel.

    :param section: Секция, в которой находится строка
    :type section: :class:`simple_report.xls.section.Section`
    :param int row_idx: Индекс строки в таблице excel
    :param int col_idx: Индекс столбца в таблице excel
    :param str text: Строка, которая будет записана в ячейку
    :param int font_size: Размер шрифта в пунктах
    :param int adjusted_row_index: Номер строки, размер которой будет увеличен.
        Если не указывается (``None``), то высота всех строк объединенной
        ячейки будет увеличена равномерно. Номер указывается относительно
        ячейки, нумерация начинается с нуля.
    """
    font = get_font(ARIAL, font_size)
    adjust_row_height(
        section, row_idx, col_idx, text, font, adjusted_row_index
    )



"Модуль с построителями отчетов"
# pylint: disable=consider-using-f-string
import os
import sys
import tempfile

from simple_report.converter.abstract import (
    FileConverter,
)
from simple_report.report import (
    DocumentReport,
    SpreadsheetReport,
)
from simple_report.xls.document import (
    DocumentXLS,
)

from . import (
    storage,
)


class SimpleReporter:
    """
    Объект занимающийся комплексным построением отчета на основе simple_report:
    1) инстанцированием и загрузкой данных провайдера
    2) инстанцированием билдера
    3) построением отчета

    Можно использовать вне паков и экшнов

    ..code:

    reporter = MySimpleReporter(provider_params, builder_params)
    report_url = reporter.make_report()
    """
    # доступные форматы
    _available_extensions = ['.xls', '.xlsx', '.docx']

    # формат по-умолчанию
    extension = '.xls'

    # путь где лежит файл шаблона отчета
    template_file_path = None
    """
    будет искать шаблон report.{extension} в директори отчета
    template_file_path = None
    custom_report.{extension} по абсолютному пути
    template_file_path = '/tmp/some/custom_report.{extension}'
    по относительному ./templates в директории отчета
    template_file_path = './templates/custom_report.{extension}'
    report.{extension} в ./templates/somedir/
    template_file_path = './templates/somedir/'
    """

    # класс провайдера для случая простого провайдера
    data_provider_class = None

    # класс билдера
    builder_class = None

    # класс адаптера
    adapter_class = None

    def __init__(self, provider_params, builder_params):
        assert self.builder_class, ''
        self.provider_params = provider_params
        self.builder_params = builder_params
        self._storage = storage.get_storage()

    def get_template(self, default_base_name='report'):
        """Возвращает путь к шаблону отчета.

        :param default_base_name: базовое имя шаблона,
            если определить не удалось
        :type default_base_name: str
        :returns: str - полный путь к шаблону
        """
        if (self.template_file_path is not None and
                os.path.isabs(self.template_file_path)):
            return self.template_file_path

        report_file_path = sys.modules[self.__module__].__file__

        report_dir = os.path.dirname(report_file_path)

        rel_sub_path = ''
        if (self.template_file_path is not None and
                self.template_file_path.startswith('./')):
            rel_sub_path = os.path.relpath(
                os.path.dirname(self.template_file_path))

        if self.template_file_path is None:
            base_name, _ = os.path.splitext(os.path.basename(report_dir))
        else:
            base_name, _ = os.path.splitext(
                os.path.basename(self.template_file_path))
            if not base_name:
                base_name = default_base_name

        auto_report_name = '{0}{1}{2}'.format(
            base_name, os.path.extsep, self.extension.strip('.'))
        auto_report_path = os.path.join(
            report_dir, rel_sub_path, auto_report_name)

        assert os.path.isfile(auto_report_path), (
            "Report template '{0}' not found at {1}".format(
                auto_report_name, auto_report_path))
        return auto_report_path

    def set_up_report(self):
        """Настройка формата отчёта."""
        template_path = self.get_template()

        if self.extension == '.xls':
            report = SpreadsheetReport(
                template_path, wrapper=DocumentXLS, type=FileConverter.XLS)
        elif self.extension == '.xlsx':
            report = SpreadsheetReport(template_path)
        elif self.extension == '.docx':
            report = DocumentReport(template_path)
        else:
            raise Exception('Unknown template extension')  # pylint: disable=broad-exception-raised

        return report

    def create_provider(self):
        """
        Кастомный метод для создания экземпляра класса провайдера.

        Используется в случае необходимости явного вызова конструктора
        провайдера, например, для композитного провайдера.
        :returns: инстанс дата-провайдера
        """

    def init_provider(self, data_provider):
        """
        Инициализирует дата-провайдер с параметрами self.provider_params
        """
        data_provider.init(**self.provider_params)

    def create_builder(self, data_provider, report):
        """
        Создание билдера.
        Если требуется, можно использовать адаптер self.adapter_class

        :returns: билдер с параметрами self.builder_params
        """
        return self.builder_class(  # pylint: disable=not-callable
            data_provider,
            adapter=self.adapter_class,
            report=report,
            params=self.builder_params
        )

    def get_data_provider(self):
        """
        Создание провайдера и взятие данных.

        :return: Provider с уже сформированным результатом
        """
        # создание провайдера
        # для случая композитного провайдера
        data_provider = self.create_provider()  # pylint: disable=assignment-from-no-return
        if data_provider is None:
            # если простой провайдер, достаточно описать его класс
            data_provider = self.data_provider_class()  # pylint: disable=not-callable

        self.init_provider(data_provider)
        data_provider.load_data()

        return data_provider

    def _get_report_builder(self, data_provider, report):
        """
        Создание билдера и построение отчета.

        :param data_provider: Provider с уже сформированным результатом.
        :param report: Отчет форматов .xls, .xlsx или .docx
        :returns: Результат выполнения метода build
        """
        # создание билдера, который будет строить кастомный отчёт
        report_builder = self.create_builder(data_provider, report=report)
        return report_builder.build()

    def build_report(self, report, params):
        """
        Построение отчета.

        :param report: Отчет форматов .xls, .xlsx или .docx
        :param out_file: str, путь к файлу.
        :param params: Результат выполение работы билдера.
        :return: None
        """
        title = self.builder_params.get('title') or 'Отчет'

        with tempfile.NamedTemporaryFile(suffix=self.extension) as fd:
            tmp_path, tmp_file = fd.name, fd.file

            if isinstance(report, DocumentReport):
                report.build(tmp_path, params)
            else:
                report.build(tmp_path)

            tmp_file.seek(0)

            return self._storage.url(
                self._storage.save(f'{title}{self.extension}', tmp_file)
            )

    def make_report(self):
        """
        Основной метод, выполняющий построение отчета
        """
        # настройка формата отчета
        report = self.set_up_report()

        # создает провайдер и формирует данные
        data_provider = self.get_data_provider()

        # создание билдера и построение отчета
        params = self._get_report_builder(data_provider, report)

        # построение отчета.
        out_url = self.build_report(report, params)

        return out_url



from datetime import (
    date,
)
from pathlib import (
    Path,
)
from unittest.case import (
    TestCase,
)

from edureporter.storage import (
    ReportsStorage,
    get_storage,
)
from testapp.app.reports.simple_report.report import (
    Reporter,
    build_report,
)


class StoragesTestCase(TestCase):
    def test_get_storage(self):
        self.assertIsInstance(get_storage(), ReportsStorage)

    def test_storage_operations(self):
        title = 'SimpleReport'

        storage = get_storage()

        path = Path(build_report(
            provider_params={'date': date.today()},
            builder_params={'title': title}
        ))

        self.assertTrue(storage.exists(path.name))
        self.assertIn(title, path.stem)
        self.assertEqual(Reporter.extension, path.suffix)

        storage.delete(path.name)

        self.assertFalse(storage.exists(path.name))



import datetime
from typing import (
    TYPE_CHECKING,
    Optional,
)

from edureporter import (
    AbstractDataProvider,
    AbstractReportBuilder,
)
from edureporter.reporter import (
    SimpleReporter,
)


if TYPE_CHECKING:
    from simple_report.report import (
        Report,
    )


class Provider(AbstractDataProvider):

    date: datetime.date

    def init(self, **params):
        self.date = params['date']

    def load_data(self):
        return {'date': self.date}


class Builder(AbstractReportBuilder):

    def __init__(self, provider: Provider, report: 'Report', *args, **kwargs):
        self._provider = provider
        self._report = report

    def build(self):
        self._flush_header()

    def _flush_header(self):
        self._flush_section('header', {
            'build_date': datetime.date.today().strftime('%d.%m.%Y'),
            'date': self._provider.date.strftime('%d.%m.%Y'),
        })

    def _flush_section(self, name: str, params: dict):
        self._report.get_section(name).flush(params)


class Reporter(SimpleReporter):

    extension = '.xlsx'
    builder_class = Builder
    data_provider_class = Provider


def build_report(
    provider_params: Optional[dict] = None,
    builder_params: Optional[dict] = None,
    **kwargs
) -> str:
    return Reporter(provider_params or {}, builder_params or {}).make_report()



# pylint: disable=import-outside-toplevel
from django.conf import (
    settings,
)
from django.core.files.storage import (
    FileSystemStorage,
)


REPORTS_STORAGE_ALIAS = 'reports'


class ReportsStorage(FileSystemStorage):

    def __init__(self, *args, location=None, base_url=None, **kwargs):
        if location is None:
            location = settings.REPORTS_DIR

        if base_url is None:
            base_url = settings.REPORTS_URL

        super().__init__(*args, location=location, base_url=base_url, **kwargs)


def get_storage():
    """Получение хранилища под разные версии Django."""
    try:
        from django.core.files.storage import (
            storages,
        )
        storage = storages[REPORTS_STORAGE_ALIAS]

    except ImportError:
        from django.core.files.storage import (
            get_storage_class,
        )
        storage = get_storage_class('edureporter.storage.ReportsStorage')()

    return storage



import os
from pathlib import (
    Path,
)

import django

from edureporter.storage import (
    REPORTS_STORAGE_ALIAS,
)


# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = Path(__file__).parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/2.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'mc^jjzr_7emr)s#k%$2_g3z_odvzj-%sye1ho42l6v_8l+(mov'

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = ['*']


# Application definition

INSTALLED_APPS = [
    'django.contrib.staticfiles',
    'testapp.app',
]

MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
]

ROOT_URLCONF = 'testapp.urls'

local_template_packages = (
    ('testapp', None),
    ('testapp', 'templates'),
)

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': list(set(
            (
                os.path.join(path, relative_path)
                if relative_path else
                os.path.dirname(path)
            )
            for name, relative_path in local_template_packages
            for path in __import__(name).__path__
        )),
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.request',
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
            ],
            'loaders': [
                ('django.template.loaders.cached.Loader', [
                    'django.template.loaders.filesystem.Loader',
                    'django.template.loaders.app_directories.Loader',
                ])
            ]
        }
    }
]


WSGI_APPLICATION = 'testapp.wsgi.application'


# Database
# https://docs.djangoproject.com/en/2.2/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.postgresql',
        'NAME': os.environ.get('POSTGRES_DB', 'testapp'),
        'USER': os.environ.get('POSTGRES_USER', 'postgres'),
        'PASSWORD': os.environ.get('POSTGRES_PASSWORD', 'postgres'),
        'HOST': os.environ.get('POSTGRES_HOST', 'localhost'),
        'PORT': 5432,
    }
}


# Internationalization
# https://docs.djangoproject.com/en/2.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/2.2/howto/static-files/

STATIC_URL = '/static/'
STATIC_ROOT = BASE_DIR / 'static'

MEDIA_URL = '/media/'
MEDIA_ROOT = BASE_DIR / 'media'

REPORTS_DIR = Path(MEDIA_ROOT) / 'downloads' / 'reports'
REPORTS_URL = f'{MEDIA_URL}downloads/reports'

if django.VERSION >= (4, 2):
    STATIC_STORAGE = {
        'BACKEND': 'django.contrib.staticfiles.storage.StaticFilesStorage',
        'OPTIONS': {
            'location': STATIC_ROOT,
            'base_url': STATIC_URL
        }
    }

    MEDIA_STORAGE = {
        'BACKEND': 'django.core.files.storage.filesystem.FileSystemStorage',
        'OPTIONS': {
            'location': MEDIA_ROOT,
            'base_url': MEDIA_URL,
        }
    }

    REPORTS_STORAGE = {
        'BACKEND': 'edureporter.storage.ReportsStorage',
        'OPTIONS': {
            'location': Path(MEDIA_ROOT) / 'downloads' / 'reports',
            'base_url': f'{MEDIA_URL}downloads/reports',
        }
    }

    STORAGES = {
        'default': MEDIA_STORAGE,
        'staticfiles': STATIC_STORAGE,
        REPORTS_STORAGE_ALIAS: REPORTS_STORAGE
    }


# Настройки Брокера ()
BROKER_URL = 'redis://localhost:6379/11'
CELERY_RESULT_BACKEND = 'redis://localhost:6379/11'
CELERYBEAT_SCHEDULER = 'django_celery_beat.schedulers.DatabaseScheduler'
CELERY_TIMEZONE = TIME_ZONE
CELERY_USE_UTC = False

LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'WARNING',
    },
    'loggers': {
        'zeep': {
            'handlers': ['console'],
            'level': 'WARNING',
            'propagate': True,
        },
    },
}



"""Шаблон для построения отчётной системы.

Представляет собой набор абстрактных базовых и вспомогательных классов для
различных элементов отчётной подсистемы: провайдера данных, построителя отчёта
и собственно отчёта.
"""
# pylint: disable=abstract-method

from abc import (
    ABCMeta,
    abstractmethod,
)
from collections.abc import (
    Mapping,
)


class AbstractDataProvider(metaclass=ABCMeta):
    """Абстрактный класс провайдера данных.

    Данный класс является базовым для всех классов, реализующих функционал
    провайдера данных. Идеология провайдера данных следующая:
        - провайдер обеспечивает доступ к данным через свойства
          (которые могут быть кешированы через декоратор @cached_property),
        - все кастомные артибуты класса, которые могут быть вызваны извне,
          при обращению к экземпляру провайдера для извлечения данных - не
          должны иметь входных параметров и желательно, должны быть объявлены
          как свойства. Кастомные атрибуты (в т.ч. методы с параметрами)
          "для внутреннего использования" должны быть приватными.
        - все необходимые исходные данные для извлечения данных, настройки
          представления и пр. - должны задаваться не в конструкторе провайдера,
          а в методе init, через словарь именованных аргументов **params,
        - провайдер должен иметь возможность вызываться отдельно, вне
          веб-интерфейса, например из Manage-команды или через Django-shell
        - провайдер не должен позволять изменять свои данные иным способом,
          кроме как последовательным вызовом методов init() и load_data().
    """

    # Имя провайдера, используемое в композитном провайдере
    provider_name = None

    def init(self, **params):
        """Инициализация провайдера.

        Используется вместо конструктора.
        """

    def load_data(self):
        """Загрузка данных.

        Данный метод может быть описан и использоваться для подготовки
        данных или реализации каких-либо необходимых общих расчётов.
        Метод не должен иметь никаких входных параметров.
        """

    @property
    def name(self):
        """Имя провайдера для применения в композитном провайдере."""
        if self.provider_name is not None:
            return self.provider_name
        return self.__class__.__name__


class CompositeDataProvider(AbstractDataProvider):
    """Композитный провайдер данных.

    Используется для объединения нескольких провайдеров в один.
    """

    def __init__(self, provider_classes):
        """Конструктор композитного провайдера.

        Здесь инстанцируются все подпровайдеры.

        :param provider_classes: список классов провайдеров
        """
        # словарь экземпляров провайдеров
        self.providers = {}

        for provider_class in provider_classes:
            if isinstance(provider_class, CompositeDataProvider):
                # пришёл инстанс композитного провайдера, такое может быть
                self.providers[provider_class.name] = provider_class
            else:
                # пришёл просто класс провайдера
                _provider = provider_class()
                self.providers[_provider.name] = _provider

    def init(self, **params):
        """Здесь инициализируются все подпровайдеры."""
        for provider in self.providers.values():
            if hasattr(provider, 'init'):
                provider.init(**params)

    def load_data(self):
        """Загрузка данных во всех подпровайдерах."""
        for provider in self.providers.values():
            if hasattr(provider, 'load_data'):
                provider.load_data()


class DependentCompositeProvider(AbstractDataProvider):
    """
    Композитный провайдер с описанием подпровайдеров, порядка их загрузки.

    'Прокидывает' данные между подпровайдерами в процессе загрузки данных
    """

    providers_order = None
    """
    Описание провайдеров в порядке их загрузки.
    Следует использовать "_" перед коротким именем провайдера,
    чтобы эти аттрибуты не уходили в адаптер

    ..code::

        providers_order = (
            ('_provider_name', ProviderClass),
            ('_provider_name2', ProviderClass2),
            ('_provider_name3', ProviderClass3),
        )
    """
    _dependence_map = None
    """
    Описание карты дополнительных данных (между подпровайдерами).
    См. метод _extend_provider_data

    ..code::

        _dependence_map = {
            '_provider_name': {
                # доп. аттрибут, который присвоится провайдеру provider_name
                'additional_provider_attribute': (
                    # у кого взять
                    _dependent_provider_name,
                    # что взять
                    dependent_provider_attribute
                ),
            }
        }
    """

    # словарь экземпляров провайдеров для поддержки работы с адаптерами
    providers = None

    def __init__(self):
        self.providers = {}
        self._providers = []
        self._state = {}
        for shortname, provider_class in self.providers_order:  # pylint: disable=not-an-iterable
            provider = provider_class()
            self._set_provider_to_loader(provider, shortname)
            # для поддержки работы с адаптерами
            self.providers[provider.name] = provider

    def _set_provider_to_loader(self, provider, shortname):
        """
        Установка зависимого подпровайдера в список для загрузки.

        :param provider: провайдер данных, наследник AbstractDataProvider
        :param str shortname: ключ из словаря self._dependence
        """
        provider._shortname = shortname  # pylint: disable=protected-access
        self._providers.append(provider)
        setattr(self, shortname, provider)

    def init(self, **params):
        """Инициализация подпровайдеров."""
        for provider in self._providers:
            if hasattr(provider, 'init'):
                provider.init(**params)

    def load_data(self):
        """Загрузка данных во всех подпровайдерах."""
        for provider in self._providers:
            if hasattr(provider, 'load_data'):
                self._extend_provider_data(provider)
                provider.load_data()

    def _extend_provider_data(self, provider):
        """
        Дополнение данных провайдера данными предыдущих загруженных провайдеров
        согласно карте зависимостей self._dependence_map

        :type provider: AbstractDataProvider
        :param provider: провайдер данных, наследник AbstractDataProvider
        """
        # дополнение параметров результатами предыдущих провайдеров
        additional = self._dependence_map.get(provider._shortname, {})  # pylint: disable=protected-access
        for param, (dependent_provider_name, att) in additional.items():
            dependent_provider = getattr(self, dependent_provider_name)
            setattr(provider, param, getattr(dependent_provider, att))


class AbstractReportBuilder(metaclass=ABCMeta):
    """Абстрактный класс построителя отчётов.

    Имеет обязательный атрибут build(), вызов которого производит
    непосредственную генерацию отчёта.
    Принимает данные от провайдера "как есть", через его свойства или целиком
    через специальный адаптер.
    """

    def __init__(self, provider=None, adapter=None, *args, **kwargs):  # pylint: disable=keyword-arg-before-vararg
        """Конструктор класса билдера.

        :param provider: Провайдер данных.
        :param adapter: Адаптер для извлечения данных из провайдера.
        """

    @abstractmethod
    def build(self):
        """ Метод, осуществляющий построение отчёта.

        Должен возвращать результат в виде файла или иной структуры данных.
        """


class BaseProviderAdapter(Mapping, dict):
    """Базовый класс для адаптеров, извлекающих данные из провайдера."""

    # список атрибутов провайдера, которые не должны быть "видны" через адаптер
    _attrs_stop_list = (
        'init',
        'load_data',
        'name',
        'provider_name',
        'providers',
    )

    def __init__(self, provider):
        """Конструктор адаптера.

        :param provider: провайдер данных
        """
        assert isinstance(provider, AbstractDataProvider), type(provider)
        self.provider = provider
        self._cache = {}

    def _check_provider_attr(self, attr):
        """Проверка атрибута - можно ли его брать из провайдера.

        'Магические' и некоторые отдельные методы будут игнорироваться.
        """
        return not(attr.startswith('_') or attr in self._attrs_stop_list)


class FlatDataProviderAdapter(BaseProviderAdapter):
    """Данные провайдера в 'плоском' виде.

    Класс адаптера, получающего все данные из провайдера и представляющего их
    в "плоском" виде (словарь с единичной глубиной):
        {
            'MainDataProvider__Provider01__param01': 16,
            'MainDataProvider__Provider01__param02': 54,
            'MainDataProvider__Provider01__param03': 80
        }
    Обращение к такому адаптеру - как и к обычному словарю, через "[]",
    """

    def __init__(self, provider, splitter):
        """Конструктор адаптера.

        :param provider: провайдер данных
        :param splitter: строка-разделитель.
        """
        assert isinstance(splitter, str), type(splitter)

        super().__init__(provider)

        self.level_splitter = splitter
        self._all_keys = self._get_all_keys(self.provider)

    def _get_all_keys(self, provider, key_level=None):
        """Все ключи в иерархии провайдеров."""
        result = []
        if key_level is None:
            key_level = ""
        else:
            if key_level:
                key_level += self.level_splitter + provider.name
            else:
                key_level = provider.name

        for attr_name in dir(provider):
            # цикл по всем атрибутам провайдера
            if not self._check_provider_attr(attr_name):
                continue

            if key_level:
                key = self.level_splitter.join([key_level, attr_name])
            else:
                key = attr_name

            result.append(key)

        if hasattr(provider, 'providers'):
            # если есть вложенные провайдеры
            for p in provider.providers.values():
                result.extend(self._get_all_keys(p, key_level))

        return result

    def __getitem__(self, item):
        """Поиск элемента с ключом item в словаре данных адаптера."""
        if not isinstance(item, str):
            raise KeyError(item)
        if item not in self._all_keys:
            raise KeyError(item)
        key_parts = item.split(self.level_splitter)
        # указатель на объект, где будет искаться атрибут
        obj = self.provider
        for part in key_parts[:-1]:
            # цикл по вложенным провайдерам
            provider = obj.providers.get(part)
            if provider:
                obj = obj.providers[provider.name]
            else:
                raise KeyError(item)
        # самый последний элемент - это уже должен быть атрибут, а не провайдер
        attr = getattr(obj, key_parts[-1])
        if callable(attr):
            # метод, свойство
            value = attr()
        else:
            # атрибут
            value = attr

        return value

    def __len__(self):
        """Магический метод для поддержки операции len()."""
        return len(self._all_keys)

    def __bool__(self):
        return bool(self._all_keys)
    __nonzero__ = __bool__

    def __iter__(self):
        return iter(self._all_keys)


class NestedDataProviderAdapter(BaseProviderAdapter):
    """Данные провайдера в 'иерархическом' (вложенные словари) виде.

    Класс адаптера, получающего все данные провайдера и представляющего их
    в виде вложенных словарей:
        {
            'MainDataProvider': {
                'Provider01': {
                    'param01': 16,
                    'param02': 54,
                    'param03': 80
                }
            }
        }
    Обращение к такому адаптеру - как и к обычному словарю, через "[]".
    """

    def __init__(self, provider):
        """Конструктор адаптера.

        :param provider: провайдер данных
        """
        super().__init__(provider)
        self._all_keys = self._get_top_keys(self.provider)

    def _get_top_keys(self, provider):
        """все необходимые атрибуты адаптируемого провайдера."""
        result = []
        for attr_name in dir(provider):
            # цикл по всем атрибутам провайдера
            if not self._check_provider_attr(attr_name):
                continue
            result.append(attr_name)

        if hasattr(provider, 'providers'):
            # если есть вложенные провайдеры
            for p in provider.providers.values():
                result.append(p.name)

        return result

    def __getitem__(self, item):
        """Поиск элемента с ключом item в словаре данных адаптера."""
        if not isinstance(item, str):
            raise KeyError(item)
        if not self._check_provider_attr(item):
            raise KeyError(item)
        if hasattr(self.provider, item):
            attr = getattr(self.provider, item)
            if callable(attr):
                value = attr()
            else:
                value = attr
            return value

        # поиск по имени провайдера
        provider = self.provider.providers.get(item)
        if provider:
            provider_key = provider.__class__.__name__
            if provider_key not in self._cache:
                self._cache[provider_key] = NestedDataProviderAdapter(provider)
            return self._cache[provider_key]
        raise KeyError(item)

    def __len__(self):
        """Магический метод для поддержки операции len()."""
        return len(self._all_keys)

    def __bool__(self):
        return bool(self._all_keys)
    __nonzero__ = __bool__

    def __iter__(self):
        return iter(self._all_keys)



"""Средства для работы со шрифтами."""
import os.path

from PIL import (
    ImageFont,
)


# Мнемоники для шрифтов
ARIAL = 1
TAHOMA = 2
CALIBRI = 3


_FONT_FILES = {
    code: os.path.join(os.path.dirname(__file__), file_name)
    for code, file_name in (
        (ARIAL, 'Arial.ttf'),
        (CALIBRI, 'Calibri.ttf'),
        (TAHOMA, 'Tahoma.ttf'),
    )
}


def get_font(font, size=10):
    """Возвращает шрифт указанного семейства.

    :param int font: Поддерживаемое модулем семейство шрифтов.
    :param int size: Размер шрифта.

    :rtype: :class:`PIL.ImageFont.FreeTypeFont`
    """
    assert font in _FONT_FILES, font

    file_path = _FONT_FILES[font]

    return ImageFont.truetype(file_path, size, encoding='utf-8')


def _split_word(font, word, max_width):
    """Разбивает слово на две части.

    Причем первая часть слова содержит наибольшее количество символов
    шрифта font от начала слова, которое умещается в max_width пикселей.

    :param font: Шрифт.
    :type font: :class:`PIL.ImageFont.FreeTypeFont`

    :param str word: Слово, которое требуется разбить.
    :param int max_width: Максимальная длина строки в пикселях.

    :rtype: tuple: (<Часть слова, умещающаяся в max_width>, <остаток>).
    """
    def _split_index(index, max_index=None):
        """Рекурсивно разбивает слово на две части.

        :param int index: Индекс текущего символа.
        :param int max_index: Индекс символа, до которого ширина строки
                              больше максимальной.
        """
        index = int(index)
        assert index > 0, ('Ширины строки не хватает, '
                           'чтобы поместить один символ')

        # определяем ширину слова в пикселях
        width, _ = font.getsize(word[:index])

        if width > max_width:

            if index == max_index:
                # Если индексы равны, значит предыдущее условие было
                # width < max_width, а предыдущий index на единицу меньше.
                return index - 1

            # Искомый index в интервале: [1; index).
            # Берем его середину.
            return _split_index(index // 2, index)

        if width < max_width:

            if max_index is None:
                # Нас обманули, слово влазит в одну строку.
                return index

            # Искомый index в интервале: [index; max_index).
            # Берем его середину.
            # Если центрального элемента нет, берем ближний справа,
            # иначе рекурсия не завершится.
            diff = max_index - index
            index += diff // 2 + diff % 2
            return _split_index(index, max_index)

        return index

    idx = _split_index(len(word))
    return word[:idx], word[idx:]


def split_text(text, font, max_width):
    """Разбивает текст на строки с учетом максимальной ширины строки.

    Разбиение осуществляется по словам. Если очередное слово не входит в
    строку, то выполняется перенос строки.

    :param unicode text: Текст, подлежащий разбиению на строки.

    :param font: Шрифт.
    :type font: :class:`PIL.ImageFont.FreeTypeFont`

    :param int max_width: Максимальная длина строки в пикселях.

    :rtype: list
    """
    words = text.split(' ')
    string = []
    strings = [string]
    while words:
        string.append(words[0])
        width, _ = font.getsize(' '.join(string))
        if width >= max_width:
            if len(string) == 1:
                word, tail = _split_word(font, words[0], max_width)
                if len(tail) > 0:
                    words[0] = tail
                else:
                    words.pop(0)
                string[0] = word
            else:
                string.pop()
            string = []
            strings.append(string)
        else:
            words.pop(0)
    return tuple(' '.join(s) for s in strings if s)
